
#import sys;sys.path.insert(1,'/home/cuvelier/Travail/Recherch/python/fc-hypermesh')
#import sys;sys.path.insert(1,'/home/cuvelier/Travail/Recherch/python/fc-tools')
#import sys;sys.path.insert(1,'/home/cuvelier/Travail/Recherch/python/fc-oogmsh')

